(* definition of lambda terms and pretty-printer *)
(* $Id: lambda.ml,v 1.51 2000/07/10 18:51:18 alamstei Exp $ *)

open Library

(* ---------------------------------------------------------------------- *)
(* Lambda terms *)

type variable = Uservar of string | Systemvar of string*int

type lambdaterm =
  | Var of variable                (* x     *)
  | App of lambdaterm * lambdaterm (* MN    *)
  | Abs of variable * lambdaterm   (* \x.M  *)
  | Let of variable * lambdaterm * lambdaterm (* let x = M in N *)
  | Letrec of variable * lambdaterm * lambdaterm (* let rec x = M in N *)
  | Intlit of int                  (* ... -1, 0, 1, ... *)
  | Boollit of bool                (* true, false *)
  | Tuple of lambdaterm list       (* (M_1, ... ,M_n) *)
  | Proj of int * int * lambdaterm (* Proj(i/n) (M_1, ..., M_n) = M_i *)
  | Inj of int * int * lambdaterm  (* inj_i/j M *)
  | Case of lambdaterm * (int*int*variable*lambdaterm) list
                                   (* Case M of inj1/n x1 -> N1 ... 
				              | injn/n xn -> Nn *)

(* some constructors *)

(* needed to insure that you always have a fresh variable for
   substitution, etc. *)
let counter = ref 0

(* uses counter function to always produce an unused variable.
   user specifies name part of variable. *)
let fresh_with_name name =
  counter := (!counter) + 1;
  Systemvar(name,!counter)

(* uses counter function to always produce an unused variable.  here the
   variable has a unique number, but no name part. *)
let fresh() = fresh_with_name ""

let rename_var var = 
  match var with
    Uservar(name) -> fresh_with_name name
  | Systemvar(name,int) -> fresh_with_name name

(* given a list of \ variables identifiers and a \ term term, return
   \ identifiers.term in the form Abs(var1, Abs(var2) ..., term) *)
let rec multiabs identifiers term = 
  match identifiers with
    hd::tl -> Abs(hd,multiabs tl term)
  | [] -> term

(* turn a lambda variable into a string *)
let string_of_variable var = 
  match var with
    Uservar(a) -> a
  | Systemvar(str,a) -> str ^"!"^ (string_of_int a)
	
(* string_of_inj i/j returns the string "inj/n".  *)
let string_of_inj j n =
  "in"^ (string_of_int j) ^"/"^ (string_of_int n)

(* puts parenthesis around string s if bool b is true *)
let optpar b s = if b then "("^s^")" else s

let rec string_of_injpattern (i,j,x,m) =
  "in"^ string_of_int i ^"/"^ string_of_int j ^" "^ 
  string_of_variable x ^" -> "^ (string_of_lambdaterm_prec 0 m)

(* Pretty-printer *)
(* This pp uses precedences to decide when to parenthesize a \-term.
   An environment has precedence 0 if no term would require parentheses.
   An environment has precedence 5 if it requires adding parentheses to
   a "let" but not to an abstraction.
   An environment has precedence 10 if it requires adding parentheses to 
   an abstraction but not an application or a projection.
   An environment has precedence 20 if it requires adding parentheses to 
   an abstraction, an application and a projection.
   Rules:
   In the term \x.M M is in an environment of precedence 0.
   In the term M N, M appears in an environment of precedence 10, and N 
   appears in an environment of precedence 20.  
   In the term Let x = M in N both M and N have a precedence of 0.
   A stand-alone term has precedence 0.  
   In the term (M1, M2, ..., Mn) each Mi has precedence 0.
   In the term Pii/j M, M has precedence 20 *)

and string_of_lambdaterm_prec prec term =
  (* Allows for arbitrarily large abstractions, i.e. terms of form 
     \a b c d ... .M *)
  let rec long_abs term' =
    match term' with
      Abs(x,m) -> " "^ (string_of_variable x) ^ (long_abs m) 
    | _ -> "." ^ (string_of_lambdaterm_prec 0 term')
  in	
  let rec vars_of_long_abs m =
    match m with
      Abs(x,m') -> " "^ (string_of_variable x) ^ (vars_of_long_abs m')
    | _ -> " = " ^ (string_of_lambdaterm_prec 5 m)
  in
  match term with
    Var(x) ->  string_of_variable x 
  | App(m,n) -> 
      optpar (prec>=20) ((string_of_lambdaterm_prec 10 m) ^" "^ 
			 (string_of_lambdaterm_prec 20 n))
  | Abs(x,m) ->
      optpar (prec>=10) ("\\" ^ (string_of_variable x) ^ (long_abs m))
  | Let(x,m,n) -> 
      optpar (prec>=5) ("let "^ (string_of_variable x) ^ 
			(vars_of_long_abs m) ^" in "^
			(string_of_lambdaterm_prec 0 n)) 
  | Letrec(x,m,n) ->
      optpar (prec>=5) ("let rec "^ (string_of_variable x) ^ 
			(vars_of_long_abs m) ^" in "^
			(string_of_lambdaterm_prec 0 n)) 
  | Intlit(n) -> string_of_int n
  | Boollit(n) -> string_of_bool n
  | Tuple(termlist) -> "("^ (Library.string_of_list 
		            (string_of_lambdaterm_prec 0) ", " termlist) ^")" 
  | Proj(i,j,term) -> optpar (prec>=20) ("pi"^ (string_of_int i) ^"/"^ 
      (string_of_int j) ^" "^ (string_of_lambdaterm_prec 20 term)) 
  | Inj(i,j,term) -> optpar (prec>=20) ("in"^ (string_of_int i) ^"/"^ 
      (string_of_int j) ^" "^ (string_of_lambdaterm_prec 20 term)) 
  | Case(term,injlist) -> optpar (prec>=5) ("case "^ 
      (string_of_lambdaterm_prec 5 term) ^" of "^ 
      (Library.string_of_list (string_of_injpattern) " | " injlist)) 

let string_of_lambdaterm term = string_of_lambdaterm_prec 0 term

(* return the free variables of the pattern injlist *)
let rec free_vars_of_injlist injlist = 
  let f (i,j,x,m) = 
    Library.set_difference (free_vars_of m) (Library.singleton x) 
  in 
  let totalfree = List.map f injlist in
  Library.set_flatten totalfree

(* return the free variables of the lambda-term term *)
and free_vars_of term = 
  match term with
    Var(x) -> Library.singleton x
  | App(m,n) ->
      Library.set_union (free_vars_of m) (free_vars_of n)
  | Abs(x,m) ->
      Library.set_difference (free_vars_of m) (Library.singleton x)
  | Let(x,m,n) -> 
      free_vars_of (App(Abs(x,n),m))
  | Letrec(x,m,n) ->
      Library.set_union (free_vars_of (Abs(x,m))) (free_vars_of (Abs(x,n)))
  | Intlit(n) -> []
  | Boollit(n) -> []
  | Tuple(termlist) -> Library.set_flatten (List.map free_vars_of termlist) 
  | Proj(i,n,m) -> free_vars_of m
  | Inj(i,j,m) -> free_vars_of m
  | Case(m,injlist) -> Library.set_union (free_vars_of m) 
	                                 (free_vars_of_injlist injlist)
 
let rec isfreein var term =
  let rec isfreein_injlist var injlist =
    match injlist with
      (i,j,x,m)::tl -> 
	if not(var=x) && isfreein var m then 
	  true 
	else 
	  (isfreein_injlist var tl)
    | [] -> false
  in	
  match term with
    Var(x) -> (x = var)
  | App(m, n) -> isfreein var m || isfreein var n
  | Abs(x, m) -> not(x = var) && isfreein var m
  | Let(x,m,n) -> isfreein var m || (not (x = var) && isfreein var n)
  | Letrec(x,m,n) -> (not (x = var)) && ((isfreein var m) || (isfreein var n)) 
  | Intlit(n) -> false
  | Boollit(n) -> false
  | Tuple(termlist) -> List.exists (isfreein var) termlist
  | Proj(i,j,m) -> isfreein var m
  | Inj(i,j,m) -> isfreein var m
  | Case(m,injlist) -> 
      (isfreein var m) || (isfreein_injlist var injlist)

(* M[x := N] *)
let rec substitute termm var termn =
  let rec substitute_injlist injlist var termn =
    match injlist with
      (i,j,x,m)::tl ->
	if (x = var) then 
	  (i,j,x,m)::(substitute_injlist tl var termn)
	else if (isfreein x termn) then
	  let x' = rename_var x in
	  (i,j,x',(substitute (substitute m x (Var x')) var termn))::
	  (substitute_injlist tl var termn)
	else
 	  (i,j,x,(substitute m var termn))::(substitute_injlist tl var termn)
    | [] -> []
  in
  match termm with
    Var(x) -> if (x = var) then termn else Var(x) 
  | App(m, n) -> App(substitute m var termn, substitute n var termn)
  | Abs(x, m) -> 
      if (x = var) then 
	Abs(x, m) 
      else if (isfreein x termn) then 
	let x' = rename_var x in
        Abs(x', substitute (substitute m x (Var x')) var termn)
      else 
	Abs(x, substitute m var termn)
  | Let(x,m,n) -> 
      let m' = substitute m var termn in
      if (x = var) then
	Let(x,m',n)
      else if (isfreein x termn) then
	let x' = rename_var x in
	Let(x',m',substitute (substitute n x (Var x')) var termn)
      else
	Let(x,m',substitute n var termn)
  | Letrec(x,m,n) ->
      if (x = var) then
	Letrec(x,m,n)
      else if (isfreein x termn) then
	let x' = rename_var x in
	let m' = substitute m x (Var x') in
      	let n' = substitute n x (Var x') in
	Letrec(x',(substitute m' var termn),(substitute n' var termn))
      else
	Letrec(x,(substitute m var termn),(substitute n var termn))
  | Intlit(n) -> Intlit(n)
  | Boollit(n) -> Boollit(n)
  | Tuple(termlist) -> 
      let termlist' =  List.map (fun m -> substitute m var termn) termlist in
      Tuple(termlist')
  | Proj(i,j,m) -> Proj(i,j,(substitute m var termn))
  | Inj(i,j,m) -> Inj(i,j,(substitute m var termn))
  | Case(m,injlist) -> 
      let m' = substitute m var termn in
      let injlist' = substitute_injlist injlist var termn in
      Case(m',injlist')
